from abc import ABC, abstractmethod
from multiprocessing import Pool
from typing import Sequence, Union

from rhoknp.units import Document, Sentence


class Processor(ABC):
    @abstractmethod
    def apply_to_document(self, document: Union[Document, str]) -> Document:
        raise NotImplementedError

    def batch_apply_to_documents(
        self, documents: Sequence[Union[Document, str]], processes: int = 0
    ) -> list[Document]:
        if processes < 1:
            return list(map(self.apply_to_document, documents))
        with Pool(processes=processes) as pool:
            return pool.map(self.apply_to_document, documents)

    @abstractmethod
    def apply_to_sentence(self, sentence: Union[Sentence, str]) -> Sentence:
        raise NotImplementedError

    def batch_apply_to_sentences(
        self, sentences: Sequence[Union[Sentence, str]], processes: int = 0
    ) -> list[Sentence]:
        if processes < 1:
            return list(map(self.apply, sentences))
        with Pool(processes=processes) as pool:
            return pool.map(self.apply, sentences)

    def apply(self, sentence: Union[Sentence, str]) -> Sentence:
        return self.apply_to_sentence(sentence)

    def batch_apply(
        self, sentences: Sequence[Union[Sentence, str]], processes: int = 0
    ) -> list[Sentence]:
        return self.batch_apply_to_sentences(sentences)



import re
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Optional, Union


def is_comment_line(line: str) -> bool:
    return line.startswith("# ") and not line.startswith("# # # 未定義語 15 その他 1 * 0 * 0")


class DepType(Enum):
    DEPENDENCY = "D"
    PARALLEL = "P"
    APPOSITION = "A"
    IMPERFECT_PARALLEL = "I"


class Semantics(dict[str, Union[str, bool]]):
    NIL = "NIL"
    PAT = re.compile(r'(?P<sems>("([^"]|\\")+?")|NIL)')
    SEM_PAT = re.compile(r"(?P<key>[^:]+)(:(?P<value>\S+))?\s?")

    def __init__(self, sstring: str):
        super().__init__()
        self.is_nil = sstring == self.NIL
        if not self.is_nil:
            for match in self.SEM_PAT.finditer(sstring.strip('"')):
                self[match.group("key")] = match.group("value") or True

    @classmethod
    def from_sstring(cls, sstring: str) -> "Semantics":
        return cls(sstring)

    def to_sstring(self) -> str:
        if self.is_nil:
            return self.NIL
        if len(self) == 0:
            return ""
        return f'"{" ".join(self._item2sem_string(k, v) for k, v in self.items())}"'

    @staticmethod
    def _item2sem_string(key: str, value: Union[str, bool]) -> str:
        if value is False:
            return ""
        if value is True:
            return f"{key}"
        return f"{key}:{value}"

    def __str__(self) -> str:
        return self.to_sstring()

    def __bool__(self) -> bool:
        return bool(dict(self)) or self.is_nil


class Features(dict[str, Union[str, bool]]):
    """A class to represent a feature information for a phrase or a base phrase

    This class parses tags in features string and converts to a dictionary.
    ex. "<正規化代表表記:遅れる/おくれる>" --> {"正規化代表表記": "遅れる/おくれる"}
    """

    PAT = re.compile(r"(?P<feats>(<[^>]+>)*)")
    TAG_PAT = re.compile(r"<(?P<key>[^:]+?)(:(?P<value>.+?))?>")

    def __init__(self, fstring: str):
        super().__init__()
        for match in self.TAG_PAT.finditer(fstring):
            self[match.group("key")] = match.group("value") or True

    @classmethod
    def from_fstring(cls, fstring: str) -> "Features":
        return cls(fstring)

    def to_fstring(self) -> str:
        return "".join(self._item2tag_string(k, v) for k, v in self.items())

    @staticmethod
    def _item2tag_string(key: str, value: Union[str, bool]) -> str:
        if value is False:
            return ""
        if value is True:
            return f"<{key}>"
        return f"<{key}:{value}>"

    def __str__(self) -> str:
        return self.to_fstring()


class RelMode(Enum):
    """同一の基本句に同一タイプの関係タグが複数付いている場合にそれらの関係を表す

        * AND: 関係の対象が並列である
            (例) 太郎と花子が学校から<帰った> (ガ格:太郎, ガ格:花子 [and])
        * OR: 「AかB」のように意味的に or である
            (例) 私は田園調布か国立に<住みたい> (ガ格:私, ニ格:田園調布, ニ格:国立 [or])
        * AMBIGUOUS:いずれの解釈も妥当であり、文脈から判断ができない
            (例) 高知県の橋本知事は…国籍条項を<撤廃する>方針を明らかにした (ガ格:高知県, ガ格:橋本知事 [？], ガ格:不特定:人 [？], ヲ格:条項, 外の関係:方針)

    Notes:
        target が「なし」の場合、同じタイプの関係タグが任意的要素であることを示す
            (例) 太郎は一人で<立っていた> (ガ格:太郎, デ格:一人, デ格:なし [？])
    """

    AND = "AND"
    OR = "OR"
    AMBIGUOUS = "？"


@dataclass
class Rel:
    PAT: ClassVar[re.Pattern[str]] = re.compile(
        r'<rel type="(?P<type>\S+?)"( mode="(?P<mode>[^>]+?)")? target="(?P<target>.+?)"( sid="(?P<sid>.+?)" '
        r'id="(?P<id>\d+?)")?/>'
    )
    type: str
    target: str
    sid: Optional[str]
    base_phrase_index: Optional[int]
    mode: Optional[RelMode]

    def to_fstring(self) -> str:
        ret = f'<rel type="{self.type}"'
        if self.mode is not None:
            ret += f' mode="{self.mode.value}"'
        ret += f' target="{self.target}"'
        if self.sid is not None:
            assert self.base_phrase_index is not None
            ret += f' sid="{self.sid}" id="{self.base_phrase_index}"'
        ret += "/>"
        return ret


class Rels(list[Rel]):
    def __init__(self, fstring: str):
        super().__init__()
        for match in Rel.PAT.finditer(fstring):
            self.append(
                Rel(
                    type=match["type"],
                    target=match["target"],
                    sid=match["sid"],
                    base_phrase_index=int(match["id"]) if match["id"] else None,
                    mode=RelMode(match["mode"]) if match["mode"] else None,
                )
            )

    @classmethod
    def from_fstring(cls, fstring: str) -> "Rels":
        return cls(fstring)

    def to_fstring(self) -> str:
        return "".join(rel.to_fstring() for rel in self)

    def __str__(self) -> str:
        return self.to_fstring()



import re
import sys
from typing import Sequence, TextIO, Union

from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.phrase import Phrase
from rhoknp.units.utils import DepType

POS_MARK = {
    "特殊": "*",
    "動詞": "v",
    "形容詞": "j",
    "判定詞": "c",
    "助動詞": "x",
    "名詞": "n",
    "固有名詞": "N",
    "人名": "J",
    "地名": "C",
    "組織名": "A",
    "指示詞": "d",
    "副詞": "a",
    "助詞": "p",
    "接続詞": "c",
    "連体詞": "m",
    "感動詞": "!",
    "接頭辞": "p",
    "接尾辞": "s",
    "未定義語": "?",
}


def draw_tree(
    leaves: Sequence[Union[Phrase, BasePhrase]],
    fh: TextIO = sys.stdout,
    show_pos: bool = True,
) -> None:
    """構文木を指定された fh に出力する．"""
    print(sprint_tree(leaves, show_pos=show_pos), file=fh, end="")


def sprint_tree(
    leaves: Sequence[Union[Phrase, BasePhrase]], show_pos: bool = True
) -> str:
    """構文木を文字列で返す．"""
    limit = len(leaves)
    item = [[""] * limit for _ in range(limit)]
    active_column = [0] * limit
    limit -= 1

    for i in range(limit):
        para_row = leaves[i].dep_type == DepType.PARALLEL
        for j in range(i + 1, limit + 1):
            if j < leaves[i].parent_index:
                if active_column[j] == 2:
                    item[i][j] = "╋" if para_row else "╂"
                elif active_column[j] == 1:
                    item[i][j] = "┿" if para_row else "┼"
                else:
                    item[i][j] = "━" if para_row else "─"
            elif j == leaves[i].parent_index:
                if leaves[i].dep_type == DepType.PARALLEL:
                    item[i][j] = "Ｐ"
                elif leaves[i].dep_type == DepType.IMPERFECT_PARALLEL:
                    item[i][j] = "Ｉ"
                elif leaves[i].dep_type == DepType.APPOSITION:
                    item[i][j] = "Ａ"
                else:
                    if active_column[j] == 2:
                        item[i][j] = "┨"
                    elif active_column[j] == 1:
                        item[i][j] = "┤"
                    else:
                        item[i][j] = "┐"
                if active_column[j] == 2:
                    # すでにＰからの太線があればそのまま
                    pass
                elif para_row:
                    active_column[j] = 2
                else:
                    active_column[j] = 1
            else:
                if active_column[j] == 2:
                    item[i][j] = "┃"
                elif active_column[j] == 1:
                    item[i][j] = "│"
                else:
                    item[i][j] = "　"

    lines = [_leaf_string(leaf, show_pos) for leaf in leaves]
    for i in range(limit):
        for j in range(i + 1, limit + 1):
            lines[i] += item[i][j]

    max_length = max(_str_real_length(line) for line in lines)
    buf = ""
    for i in range(limit + 1):
        diff = max_length - _str_real_length(lines[i])
        buf += " " * diff
        buf += lines[i] + "\n"

    return buf


def _leaf_string(leaf: Union[Phrase, BasePhrase], show_pos: bool) -> str:
    ret = ""
    for mrph in leaf.morphemes:
        ret += mrph.text
        if show_pos is True:
            if re.search("^(?:固有名詞|人名|地名)$", mrph.subpos):
                ret += POS_MARK[mrph.subpos]
            else:
                ret += POS_MARK[mrph.pos]
    return ret


def _str_real_length(string: str) -> int:
    length = 0
    for char in string:
        if re.search(r"^[a-zA-Z*!?]$", char):
            # 品詞情報は長さ1
            length += 1
        else:
            length += 2
    return length



import weakref
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from rhoknp.pas.pas import Pas
    from rhoknp.units.base_phrase import BasePhrase


class Predicate:
    def __init__(self, unit: "BasePhrase", cfid: Optional[str] = None):
        self.unit: "BasePhrase" = unit
        self.cfid = cfid
        self._pas: Optional["Pas"] = None

    @property
    def phrase(self) -> "BasePhrase":
        return self.unit

    @property
    def text(self) -> str:
        return self.unit.text

    @property
    def pas(self) -> "Pas":
        if self._pas is None:
            raise AttributeError("pas has not been set")
        return self._pas

    @pas.setter
    def pas(self, pas: "Pas") -> None:
        self._pas = weakref.proxy(pas)

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text={repr(self.text)})"



from importlib.metadata import version

from rhoknp.processors import KNP, Jumanpp, RegexSenter
from rhoknp.units import BasePhrase, Clause, Document, Morpheme, Phrase, Sentence

__version__ = version("rhoknp")

__all__ = [
    "__version__",
    # processors
    "RegexSenter",
    "Jumanpp",
    "KNP",
    # units
    "Document",
    "Sentence",
    "Clause",
    "Phrase",
    "BasePhrase",
    "Morpheme",
]



import weakref
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.phrase import Phrase
from rhoknp.units.unit import Unit

if TYPE_CHECKING:
    from rhoknp.units.document import Document
    from rhoknp.units.sentence import Sentence


class Clause(Unit):
    """節クラス．"""

    count = 0

    def __init__(self) -> None:
        super().__init__()

        # parent unit
        self._sentence: Optional["Sentence"] = None

        # child units
        self._phrases: Optional[list[Phrase]] = None

        self.index = self.count
        Clause.count += 1

    @property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        offset = 0
        for prev_sentence in self.document.sentences[: self.sentence.index]:
            offset += len(prev_sentence.clauses)
        return self.index + offset

    @property
    def parent_unit(self) -> Optional["Sentence"]:
        """上位の言語単位（文）．未登録なら None．"""
        return self._sentence

    @property
    def child_units(self) -> Optional[list[Phrase]]:
        """下位の言語単位（文節）．解析結果にアクセスできないなら None．"""
        return self._phrases

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.sentence.document

    @property
    def sentence(self) -> "Sentence":
        """文．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self.parent_unit is None:
            raise AttributeError("sentence has not been set")
        return self.parent_unit

    @sentence.setter
    def sentence(self, sentence: "Sentence") -> None:
        """文．

        Args:
            sentence: 文．
        """
        self._sentence = sentence

    @property
    def phrases(self) -> list[Phrase]:
        """文節のリスト．"""
        if self._phrases is None:
            raise AssertionError
        return self._phrases

    @phrases.setter
    def phrases(self, phrases: list[Phrase]) -> None:
        """文節のリスト．

        Args:
            phrases: 文節のリスト．
        """
        for phrase in phrases:
            phrase.clause = weakref.proxy(self)
        self._phrases = phrases

    @property
    def base_phrases(self) -> list[BasePhrase]:
        """基本句のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [
            base_phrase
            for phrase in self.phrases
            for base_phrase in phrase.base_phrases
        ]

    @property
    def morphemes(self) -> list[Morpheme]:
        """形態素のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [
            morpheme
            for base_phrase in self.base_phrases
            for morpheme in base_phrase.morphemes
        ]

    @cached_property
    def head(self) -> BasePhrase:
        """節主辞の基本句．"""
        for base_phrase in self.base_phrases:
            if base_phrase.features and "節-主辞" in base_phrase.features:
                return base_phrase
        raise AssertionError

    @cached_property
    def parent(self) -> Optional["Clause"]:
        """係り先の節．ないなら None．"""
        head_parent = self.head.parent
        while head_parent in self.base_phrases:
            head_parent = head_parent.parent
        for clause in self.sentence.clauses:
            if head_parent in clause.base_phrases:
                return clause
        return None

    @cached_property
    def children(self) -> list["Clause"]:
        """この節に係っている節のリスト．"""
        return [clause for clause in self.sentence.clauses if clause.parent == self]

    @classmethod
    def from_knp(cls, knp_text: str) -> "Clause":
        """節クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．
        """
        clause = cls()
        phrases = []
        phrase_lines: list[str] = []
        for line in knp_text.split("\n"):
            if not line.strip():
                continue
            if line.startswith("*") and phrase_lines:
                phrases.append(Phrase.from_knp("\n".join(phrase_lines)))
                phrase_lines = []
            phrase_lines.append(line)
        else:
            phrase = Phrase.from_knp("\n".join(phrase_lines))
            phrases.append(phrase)
        clause.phrases = phrases
        return clause

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        return "".join(phrase.to_knp() for phrase in self.phrases)



import weakref
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional

from rhoknp.units.base_phrase import BasePhrase

if TYPE_CHECKING:
    from rhoknp.pas.pas import Pas
    from rhoknp.units.clause import Clause
    from rhoknp.units.document import Document
    from rhoknp.units.phrase import Phrase
    from rhoknp.units.sentence import Sentence


class ArgumentType(Enum):
    """
    ref: https://nlp.ist.i.kyoto-u.ac.jp/index.php?KNP%2F%E6%A0%BC%E8%A7%A3%E6%9E%90%E7%B5%90%E6%9E%9C%E6%9B%B8%E5%BC%8F
    """

    CASE_EXPLICIT = "C"  # a.k.a. overt
    CASE_HIDDEN = "N"
    OMISSION = "O"
    DEMONSTRATIVE = "D"
    EXOPHOR = "E"
    UNASSIGNED = "U"


class BaseArgument(ABC):
    """A base class for all kinds of arguments"""

    def __init__(self, arg_type: ArgumentType):
        self.type: ArgumentType = arg_type
        self.optional = False
        self._pas: Optional["Pas"] = None

    @property
    def is_special(self) -> bool:
        return self.type == ArgumentType.EXOPHOR

    @property
    def pas(self) -> "Pas":
        if self._pas is None:
            raise AttributeError("pas has not been set")
        return self._pas

    @pas.setter
    def pas(self, pas: "Pas") -> None:
        self._pas = weakref.proxy(pas)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        raise NotImplementedError


class Argument(BaseArgument):
    def __init__(self, base_phrase: BasePhrase, arg_type: ArgumentType):
        super().__init__(arg_type)
        self.base_phrase: BasePhrase = base_phrase

    @property
    def unit(self) -> BasePhrase:
        return self.base_phrase

    @property
    def document(self) -> "Document":
        return self.base_phrase.document

    @property
    def sentence(self) -> "Sentence":
        return self.base_phrase.sentence

    @property
    def clause(self) -> "Clause":
        return self.base_phrase.clause

    @property
    def phrase(self) -> "Phrase":
        return self.base_phrase.phrase

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_phrase={repr(self.base_phrase)}, arg_type={repr(self.type)})"

    def __str__(self) -> str:
        return self.base_phrase.text

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Argument) and self.base_phrase == other.base_phrase


class SpecialArgument(BaseArgument):
    """外界を指す項を表すクラス

    Args:
        exophor (str): 外界照応詞 (不特定:人など)
        eid (int): 外界照応詞のエンティティID
    """

    def __init__(self, exophor: str, eid: int):
        super().__init__(ArgumentType.EXOPHOR)
        self.exophor: str = exophor
        self.eid: int = eid

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(exophor={repr(self.exophor)}, eid={repr(self.eid)})"

    def __str__(self) -> str:
        return self.exophor

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SpecialArgument) and self.exophor == other.exophor



from logging import getLogger
from subprocess import PIPE, Popen, run
from typing import Optional, Union

from rhoknp.processors.processor import Processor
from rhoknp.processors.senter import RegexSenter
from rhoknp.units import Document, Sentence

logger = getLogger(__file__)


class Jumanpp(Processor):
    """Jumanpp クラス．

    Args:
        executable: Juman++ のパス．
        options: Juman++ のオプション．
        senter: 文分割器のインスタンス．文分割がまだなら，先にこのインスタンスを用いて文分割する．
            未設定なら RegexSenter を使って文分割する．

    Example::

        from rhoknp import Jumanpp

        jumanpp = Jumanpp()
        sentence = jumanpp.apply("電気抵抗率は、どんな材料が電気を通しにくいかを比較するために、用いられる物性値である。")
    """

    def __init__(
        self,
        executable: str = "jumanpp",
        options: Optional[list[str]] = None,
        senter: Optional[Processor] = None,
    ):
        self.executable = executable
        self.options = options

        self.senter = senter

    def __repr__(self) -> str:
        arg_string = f"executable={repr(self.executable)}"
        if self.options is not None:
            arg_string += f", options={repr(self.options)}"
        if self.senter is not None:
            arg_string += f", senter={repr(self.senter)}"
        return f"{self.__class__.__name__}({arg_string})"

    def apply_to_document(self, document: Union[Document, str]) -> Document:
        """文書に Jumanpp を適用する．

        Args:
            document: 文書．

        .. note::
            文分割がまだなら，先に初期化時に設定した senter で文分割する．
            未設定なら RegexSenter で文分割する．
        """
        if isinstance(document, str):
            document = Document(document)

        if document.need_senter:
            logger.debug("document needs to be split into sentences")
            if self.senter is None:
                logger.debug(
                    "senter is not specified when initializing KNP: use RegexSenter with no option"
                )
                self.senter = RegexSenter()
            document = self.senter.apply_to_document(document)

        with Popen(self.run_command, stdout=PIPE, stdin=PIPE, encoding="utf-8") as p:
            jumanpp_text, _ = p.communicate(input=document.to_plain())
        return Document.from_jumanpp(jumanpp_text)

    def apply_to_sentence(self, sentence: Union[Sentence, str]) -> Sentence:
        """文に Jumanpp を適用する．

        Args:
            sentence: 文．
        """
        if isinstance(sentence, str):
            sentence = Sentence(sentence)

        with Popen(self.run_command, stdout=PIPE, stdin=PIPE, encoding="utf-8") as p:
            jumanpp_text, _ = p.communicate(input=sentence.to_plain())
        return Sentence.from_jumanpp(jumanpp_text)

    def is_available(self) -> bool:
        """Jumanpp が利用可能であれば True を返す．"""
        try:
            p = run(self.version_command, stdout=PIPE, stdin=PIPE, encoding="utf-8")
            logger.info(p.stdout.strip())
            return True
        except Exception as e:
            logger.warning(e)
            return False

    @property
    def run_command(self) -> list[str]:
        """解析時に実行するコマンド．"""
        command = [self.executable]
        if self.options:
            command += self.options
        return command

    @property
    def version_command(self) -> list[str]:
        """バージョン確認時に実行するコマンド．"""
        return [self.executable, "-v"]



from rhoknp.pas.argument import Argument, ArgumentType, SpecialArgument
from rhoknp.pas.pas import Pas
from rhoknp.pas.predicate import Predicate

__all__ = ["Predicate", "ArgumentType", "Argument", "SpecialArgument", "Pas"]



import re
import weakref
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from rhoknp.units.morpheme import Morpheme
from rhoknp.units.unit import Unit
from rhoknp.units.utils import DepType, Features, Rels

if TYPE_CHECKING:
    from rhoknp.pas.pas import Pas
    from rhoknp.units.clause import Clause
    from rhoknp.units.document import Document
    from rhoknp.units.phrase import Phrase
    from rhoknp.units.sentence import Sentence


class BasePhrase(Unit):
    """基本句クラス．"""

    KNP_PAT = re.compile(
        rf"^\+ (?P<pid>-1|\d+)(?P<dtype>[{''.join(e.value for e in DepType)}])( (?P<tags>(<[^>]+>)*))?$"
    )
    count = 0

    def __init__(
        self, parent_index: int, dep_type: DepType, features: Features, rels: Rels
    ):
        super().__init__()

        # parent unit
        self._phrase: Optional["Phrase"] = None

        # child units
        self._morphemes: Optional[list[Morpheme]] = None

        self.parent_index: int = parent_index  #: 係り先の基本句の文内におけるインデックス．
        self.dep_type: DepType = dep_type  #: 係り受けの種類．
        self.features: Features = features  #: 素性．
        self.rels: Rels = rels  #: 基本句間関係．

        self.index = self.count
        BasePhrase.count += 1

        # Predicate-argument structure
        self._pas: Optional["Pas"] = None

    @property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        offset = 0
        for prev_sentence in self.document.sentences[: self.sentence.index]:
            offset += len(prev_sentence.base_phrases)
        return self.index + offset

    @property
    def parent_unit(self) -> Optional["Phrase"]:
        """上位の言語単位（文節）．未登録なら None．"""
        return self._phrase

    @property
    def child_units(self) -> Optional[list[Morpheme]]:
        """下位の言語単位（形態素）．解析結果にアクセスできないなら None．"""
        return self._morphemes

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.document

    @property
    def sentence(self) -> "Sentence":
        """文．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.sentence

    @property
    def clause(self) -> "Clause":
        """節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.clause

    @property
    def phrase(self) -> "Phrase":
        """文節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._phrase is None:
            raise AttributeError("phrase has not been set")
        return self._phrase

    @phrase.setter
    def phrase(self, phrase: "Phrase") -> None:
        """文節．

        Args:
            phrase: 文節．
        """
        self._phrase = phrase

    @property
    def morphemes(self) -> list[Morpheme]:
        """形態素．"""
        if self._morphemes is None:
            raise AssertionError
        return self._morphemes

    @morphemes.setter
    def morphemes(self, morphemes: list[Morpheme]) -> None:
        """形態素．

        Args:
            morphemes: 形態素．
        """
        for morpheme in morphemes:
            morpheme.base_phrase = weakref.proxy(self)
        self._morphemes = morphemes

    @cached_property
    def head(self) -> Morpheme:
        """主辞の形態素．"""
        head = None
        for morpheme in self.morphemes:
            if morpheme.features is None:
                continue
            if "内容語" in morpheme.features and head is None:
                # Consider the first content word as the head
                head = morpheme
            if "準内容語" in morpheme.features:
                # Sub-content words overwrite the head
                head = morpheme
        if head:
            return head
        return self.morphemes[0]

    @property
    def parent(self) -> Optional["BasePhrase"]:
        """係り先の基本句．ないなら None．"""
        if self.parent_index == -1:
            return None
        return self.sentence.base_phrases[self.parent_index]

    @cached_property
    def children(self) -> list["BasePhrase"]:
        """この基本句に係っている基本句のリスト．"""
        return [
            base_phrase
            for base_phrase in self.sentence.base_phrases
            if base_phrase.parent == self
        ]

    @property
    def pas(self) -> "Pas":
        """述語項構造．"""
        if self._pas is None:
            raise AttributeError("pas has not been set")
        return self._pas

    @pas.setter
    def pas(self, pas: "Pas") -> None:
        """述語項構造．

        Args:
            pas: 述語項構造．
        """
        self._pas = weakref.proxy(pas)

    @classmethod
    def from_knp(cls, knp_text: str) -> "BasePhrase":
        """基本句クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．
        """
        first_line, *lines = knp_text.split("\n")
        match = cls.KNP_PAT.match(first_line)
        if match is None:
            raise ValueError(f"malformed line: {first_line}")
        parent_index = int(match.group("pid"))
        dep_type = DepType(match.group("dtype"))
        features = Features(match.group("tags") or "")
        rels = Rels.from_fstring(match.group("tags") or "")
        base_phrase = cls(parent_index, dep_type, features, rels)

        morphemes: list[Morpheme] = []
        for line in lines:
            if not line.strip():
                continue
            morphemes.append(Morpheme.from_jumanpp(line))
        base_phrase.morphemes = morphemes
        return base_phrase

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        ret = f"+ {self.parent_index}{self.dep_type.value}"
        if self.features:
            ret += f" {self.features}"
        ret += "\n"
        ret += "".join(morpheme.to_jumanpp() for morpheme in self.morphemes)
        return ret



import re
from dataclasses import astuple, dataclass, fields
from functools import cached_property
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from rhoknp.units.unit import Unit
from rhoknp.units.utils import Features, Semantics

if TYPE_CHECKING:
    from rhoknp.units.base_phrase import BasePhrase
    from rhoknp.units.clause import Clause
    from rhoknp.units.document import Document
    from rhoknp.units.phrase import Phrase
    from rhoknp.units.sentence import Sentence


@dataclass(frozen=True)
class MorphemeAttributes:
    """形態素の属性クラス．"""

    JUMANPP_PAT = re.compile(
        r"(?P<attrs>([^ ]+ [^ ]+ [^ ]+ [^ ]+ \d+ [^ ]+ \d+ [^ ]+ \d+ [^ ]+ \d+))"
    )

    surf: str  #: 表層表現．
    reading: str  #: 読み．
    lemma: str  #: 原形．
    pos: str  #: 品詞．
    pos_id: int  #: 品詞ID．
    subpos: str  #: 品詞細分類．
    subpos_id: int  #: 品詞細分類ID．
    conjtype: str  #: 活用型．
    conjtype_id: int  #: 活用型ID．
    conjform: str  #: 活用形ID．
    conjform_id: int  #: 活用形ID．

    @classmethod
    def from_jumanpp(cls, jumanpp_line: str) -> "MorphemeAttributes":
        """形態素の属性クラスのインスタンスを Juman++ の解析結果から初期化．

        Args:
            jumanpp_line: Juman++ の解析結果．
        """
        kwargs = {}
        for field, value in zip(fields(cls), jumanpp_line.split(" ")):
            kwargs[field.name] = field.type(value)
        assert len(kwargs) == len(fields(cls)), f"malformed line: {jumanpp_line}"
        return cls(**kwargs)

    def to_jumanpp(self) -> str:
        """Juman++ フォーマットに変換．"""
        return " ".join(str(item) for item in astuple(self))


class Morpheme(Unit):
    """形態素クラス．"""

    JUMANPP_PAT: ClassVar[re.Pattern[str]] = re.compile(
        (
            rf"^({MorphemeAttributes.JUMANPP_PAT.pattern})"
            + rf"(\s{Semantics.PAT.pattern})?"
            + rf"(\s{Features.PAT.pattern})?$"
        )
    )

    count = 0

    def __init__(
        self,
        attributes: MorphemeAttributes,
        semantics: Semantics,
        features: Features,
        homograph: bool = False,
    ):
        super().__init__()

        # parent unit
        self._base_phrase: Optional["BasePhrase"] = None
        self._sentence: Optional["Sentence"] = None

        self._attributes = attributes
        self.semantics = semantics  #: 辞書 (JumanDic) に記載の意味情報．
        self.features = features  #: 素性．
        self.homographs: list["Morpheme"] = []  #: 同形の形態素のリスト．

        self.text = attributes.surf

        self.index = self.count
        if homograph is False:
            Morpheme.count += 1

    @property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        offset = 0
        for prev_sentence in self.document.sentences[: self.sentence.index]:
            offset += len(prev_sentence.morphemes)
        return self.index + offset

    @property
    def parent_unit(self) -> Optional[Union["BasePhrase", "Sentence"]]:
        """上位の言語単位（基本句もしくは文）．未登録なら None．

        ..note::
            KNP によって解析済みなら基本句， Jumanpp によって解析済みなら文を返却．
        """
        if self._base_phrase is not None:
            return self._base_phrase
        if self._sentence is not None:
            return self._sentence
        return None

    @property
    def child_units(self) -> None:
        """下位の言語単位のリスト．形態素は最下位の言語単位なので常に None．"""
        return None

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.sentence.document

    @property
    def sentence(self) -> "Sentence":
        """文．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._sentence is not None:
            return self._sentence
        if self._base_phrase is not None:
            return self.clause.sentence
        raise AttributeError("sentence has not been set")

    @sentence.setter
    def sentence(self, sentence: "Sentence") -> None:
        """文．

        Args:
            sentence: 文．
        """
        self._sentence = sentence

    @property
    def clause(self) -> "Clause":
        """節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.phrase.clause

    @property
    def phrase(self) -> "Phrase":
        """文節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.base_phrase.phrase

    @property
    def base_phrase(self) -> "BasePhrase":
        """基本句．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._base_phrase is None:
            raise AttributeError("not available before applying KNP")
        return self._base_phrase

    @base_phrase.setter
    def base_phrase(self, base_phrase: "BasePhrase") -> None:
        """基本句．

        Args:
            base_phrase: 基本句．
        """
        self._base_phrase = base_phrase

    @property
    def surf(self) -> str:
        """表層表現．"""
        return self._attributes.surf

    @property
    def reading(self) -> str:
        """読み．"""
        return self._attributes.reading

    @property
    def lemma(self) -> str:
        """原形．"""
        return self._attributes.lemma

    @property
    def canon(self) -> Optional[str]:
        """代表表記．"""
        canon = self.semantics.get("代表表記", None)
        assert canon is None or isinstance(canon, str)
        return canon

    @property
    def pos(self) -> str:
        """品詞．"""
        return self._attributes.pos

    @property
    def subpos(self) -> str:
        """品詞細分類．"""
        return self._attributes.subpos

    @property
    def conjtype(self) -> str:
        """活用型．"""
        return self._attributes.conjtype

    @property
    def conjform(self) -> str:
        """活用形．"""
        return self._attributes.conjform

    @property
    def sstring(self) -> str:
        """Juman++ フォーマットの意味情報．"""
        return self.semantics.to_sstring()

    @property
    def fstring(self) -> str:
        """Juman++ フォーマットの素性．"""
        return self.features.to_fstring()

    @cached_property
    def parent(self) -> Optional["Morpheme"]:
        """係り先の形態素．ないなら None．"""
        if self.base_phrase.head == self:
            if self.base_phrase.parent is not None:
                return self.base_phrase.parent.head
            return None
        return self.base_phrase.head

    @cached_property
    def span(self) -> tuple[int, int]:
        """文中での文字レベルのスパン．"""
        if self._sentence is None or self.index == 0:
            start = 0
        else:
            _, prev_end = self.sentence.morphemes[self.index - 1].span
            start = prev_end
        end = start + len(self.text)  # TODO: correctly handle multibyte characters
        return start, end

    @cached_property
    def children(self) -> list["Morpheme"]:
        """この形態素に係っている形態素のリスト．"""
        return [
            morpheme for morpheme in self.sentence.morphemes if morpheme.parent == self
        ]

    @classmethod
    def from_jumanpp(cls, jumanpp_text: str) -> "Morpheme":
        """形態素クラスのインスタンスを Juman++ の解析結果から初期化．

        Args:
            jumanpp_text: Juman++ の解析結果．
        """
        first_line, *lines = jumanpp_text.rstrip().split("\n")
        morpheme = cls._from_jumanpp_line(first_line)
        for line in lines:
            assert line.startswith("@ ")
            homograph = cls._from_jumanpp_line(line[2:], homograph=True)
            morpheme.homographs.append(homograph)
        return morpheme

    @classmethod
    def _from_jumanpp_line(
        cls, jumanpp_line: str, homograph: bool = False
    ) -> "Morpheme":
        """形態素クラスのインスタンスを Juman++ の解析結果から初期化．

        Args:
            jumanpp_line: Juman++ の解析結果．
            homograph: 同形かどうかを表すフラグ．
        """
        assert "\n" not in jumanpp_line.strip()
        match = cls.JUMANPP_PAT.match(jumanpp_line)
        if match is None:
            raise ValueError(f"malformed line: {jumanpp_line}")
        attributes = MorphemeAttributes.from_jumanpp(match.group("attrs"))
        semantics = Semantics.from_sstring(match.group("sems") or "")
        features = Features.from_fstring(match.group("feats") or "")
        return cls(attributes, semantics, features, homograph=homograph)

    def to_jumanpp(self) -> str:
        """Juman++ フォーマットに変換．"""
        ret = self._attributes.to_jumanpp()
        if self.semantics:
            ret += f" {self.semantics}"
        if self.features:
            ret += f" {self.features}"
        ret += "\n"
        for homograph in self.homographs:
            ret += f"@ {homograph.to_jumanpp()}"
        return ret



import weakref
from logging import getLogger
from typing import Optional, Sequence, Union

from rhoknp.pas.pas import Pas
from rhoknp.pas.predicate import Predicate
from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.clause import Clause
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.phrase import Phrase
from rhoknp.units.sentence import Sentence
from rhoknp.units.unit import Unit
from rhoknp.units.utils import is_comment_line
from rhoknp.utils.constants import ALL_CASES

logger = getLogger(__file__)


class Document(Unit):
    """文書クラス．

    Args:
        text: 文書の文字列．
    """

    count = 0

    def __init__(self, text: Optional[str] = None):
        super().__init__()

        Sentence.count = 0

        # child units
        self._sentences: Optional[list[Sentence]] = None

        if text is not None:
            self.text = text

        self.index = self.count
        Document.count += 1

        self.has_error = False

        self._pass: list[Pas] = []

    def _post_init(self) -> None:
        """インスタンス作成後の追加処理を行う．"""
        if self.need_knp is False:
            self._parse_rel()

    @property
    def parent_unit(self) -> None:
        """上位の言語単位．文書は最上位の言語単位なので常に None．"""
        return None

    @property
    def child_units(self) -> Optional[list[Sentence]]:
        """下位の言語単位（文）のリスト．解析結果にアクセスできないなら None．"""
        return self._sentences

    @property
    def sentences(self) -> list[Sentence]:
        """文のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._sentences is None:
            raise AttributeError("not available before applying a sentence splitter")
        return self._sentences

    @sentences.setter
    def sentences(self, sentences: list[Sentence]) -> None:
        """文のリスト．

        Args:
            sentences: 文のリスト．
        """
        for sentence in sentences:
            sentence.document = weakref.proxy(self)
        self._sentences = sentences

    @property
    def clauses(self) -> list[Clause]:
        """節のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [clause for sentence in self.sentences for clause in sentence.clauses]

    @property
    def phrases(self) -> list[Phrase]:
        """文節のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [phrase for sentence in self.sentences for phrase in sentence.phrases]

    @property
    def base_phrases(self) -> list[BasePhrase]:
        """基本句のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [
            base_phrase
            for sentence in self.sentences
            for base_phrase in sentence.base_phrases
        ]

    @property
    def morphemes(self) -> list[Morpheme]:
        """形態素のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [
            morpheme for sentence in self.sentences for morpheme in sentence.morphemes
        ]

    @property
    def need_senter(self) -> bool:
        """文分割がまだなら True．"""
        return self._sentences is None

    @property
    def need_jumanpp(self) -> bool:
        """Juman++ による形態素解析がまだなら True．"""
        return self.need_senter or any(
            sentence.need_jumanpp for sentence in self.sentences
        )

    @property
    def need_knp(self) -> bool:
        """KNP による構文解析がまだなら True．"""
        return self.need_senter or any(sentence.need_knp for sentence in self.sentences)

    def pas_list(self) -> list[Pas]:
        """述語項構造のリストを返却．"""
        return self._pass

    @classmethod
    def from_raw_text(cls, text: str) -> "Document":
        """文書クラスのインスタンスを文書の生テキストから初期化．

        Args:
            text: 文書の生テキスト．

        Example::

            from rhoknp import Document

            doc_text = "天気が良かったので散歩した。途中で先生に会った。"
            doc = Document.from_raw_text(doc_text)
        """
        document = cls(text)
        document._post_init()
        return document

    @classmethod
    def from_line_by_line_text(cls, text: str) -> "Document":
        """文書クラスのインスタンスを一行一文形式のテキストから初期化．

        Args:
            text: 一行一文形式に整形された文書のテキスト．

        Example::

            from rhoknp import Document

            sent_texts = \"\"\"# S-ID:1
            天気が良かったので散歩した。
            # S-ID:2
            途中で先生に会った。
            \"\"\"
            doc = Document.from_line_by_line_text(sent_texts)

        .. note::
            # から始まる行は直後の文に対するコメントとして認識される．
        """
        document = cls()
        sentences = []
        sentence_lines: list[str] = []
        for line in text.split("\n"):
            sentence_lines.append(line)
            if is_comment_line(line):
                continue
            sentences.append(Sentence.from_raw_text("\n".join(sentence_lines)))
            sentence_lines = []
        document.sentences = sentences
        document._post_init()
        return document

    @classmethod
    def from_sentences(
        cls, sentences: Union[Sequence[Union[Sentence, str]], str]
    ) -> "Document":
        """文書クラスのインスタンスを文のリストから初期化．

        Args:
            sentences: 文（文の文字列）のリスト．

        Example::

            from rhoknp import Document

            sent_texts = ["天気が良かったので散歩した。", "途中で先生に会った。"]
            doc = Document.from_sentences(sent_texts)
        """
        document = cls()
        sentences_ = []
        for sentence in sentences:
            if isinstance(sentence, Sentence):
                sentence = sentence.text
            sentences_.append(Sentence.from_raw_text(sentence))
        document.sentences = sentences_
        document._post_init()
        return document

    @classmethod
    def from_jumanpp(cls, jumanpp_text: str, ignore_errors: bool = False) -> "Document":
        """文書クラスのインスタンスを Juman++ の解析結果から初期化．

        Args:
            jumanpp_text: Juman++ の解析結果．
            ignore_errors: 解析結果中にエラーが発生してもその文を捨てて処理を続行する．

        Example::

            from rhoknp import Document

            jumanpp_text = \"\"\"天気 てんき 天気 名詞 6 普通名詞 1 * 0 * 0 "代表表記:天気/てんき カテゴリ:抽象物" <代表表記:天気/てんき><カテゴリ:抽象物><正規化代表表記:天気/てんき><漢字><かな漢字><名詞相当語><文頭><自立><内容語><タグ単位始><文節始><文節主辞>
            が が が 助詞 9 格助詞 1 * 0 * 0 NIL <かな漢字><ひらがな><付属>
            良かった よかった 良い 形容詞 3 * 0 イ形容詞アウオ段 18 タ形 8 "代表表記:良い/よい 反義:形容詞:悪い/わるい"
            ので ので のだ 助動詞 5 * 0 ナ形容詞 21 ダ列タ系連用テ形 12 NIL <かな漢字><ひらがな><活用語><付属>
            散歩 さんぽ 散歩 名詞 6 サ変名詞 2 * 0 * 0 "代表表記:散歩/さんぽ ドメイン:レクリエーション カテゴリ:抽象物" <代表表記:散歩/さんぽ><ドメイン:レクリエーション><カテゴリ:抽象物><正規化代表表記:散歩/さんぽ><漢字><かな漢字><名詞相当語><サ変><サ変動詞><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            した した する 動詞 2 * 0 サ変動詞 16 タ形 10 "代表表記:する/する 自他動詞:自:成る/なる 付属動詞候補（基本）" <代表表記:する/する><自他動詞:自:成る/なる><付属動詞候補（基本）><正規化代表表記:する/する><かな漢字><ひらがな><活用語><表現文末><とタ系連用テ形複合辞><付属>
            。 。 。 特殊 1 句点 1 * 0 * 0 NIL <英記号><記号><文末><付属>
            EOS
            途中 とちゅう 途中 名詞 6 時相名詞 10 * 0 * 0 "代表表記:途中/とちゅう カテゴリ:抽象物 弱時相名詞 修飾（デ格）"
            で で で 助詞 9 格助詞 1 * 0 * 0 NIL
            先生 せんせい 先生 名詞 6 普通名詞 1 * 0 * 0 "代表表記:先生/せんせい ドメイン:教育・学習 カテゴリ:人 人名末尾"
            に に に 助詞 9 格助詞 1 * 0 * 0 NIL
            会った あった 会う 動詞 2 * 0 子音動詞ワ行 12 タ形 10 "代表表記:会う/あう 反義:動詞:分かれる/わかれる;動詞:別れる/わかれる"
            EOS
            \"\"\"
            doc = Document.from_jumanpp(jumanpp_text)

        .. note::
            複数文の解析結果が含まれている場合，一つの文書として扱われる．．
        """
        document = cls()
        sentences = []
        sentence_lines: list[str] = []
        for line in jumanpp_text.split("\n"):
            if line.strip() == "":
                continue
            sentence_lines.append(line)
            if line.strip() == Sentence.EOS:
                try:
                    sentences.append(
                        Sentence.from_jumanpp("\n".join(sentence_lines) + "\n")
                    )
                except Exception as e:
                    document.has_error = True
                    if not ignore_errors:
                        raise e
                sentence_lines = []
        document.sentences = sentences
        document._post_init()
        return document

    @classmethod
    def from_knp(cls, knp_text: str, ignore_errors: bool = False) -> "Document":
        """文書クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．
            ignore_errors: 解析結果中にエラーが発生してもその文を捨てて処理を続行する．

        Example::

            from rhoknp import Document

            knp_text = \"\"\"# S-ID: 1
            * 1D <BGH:天気/てんき><文頭><ガ><助詞><体言><係:ガ格><区切:0-0><格要素><連用要素><正規化代表表記:天気/てんき><主辞代表表記:天気/てんき>
            + 1D <BGH:天気/てんき><文頭><ガ><助詞><体言><係:ガ格><区切:0-0><格要素><連用要素><名詞項候補><先行詞候補><正規化代表表記:天気/てんき><主辞代表表記:天気/てんき>
            天気 てんき 天気 名詞 6 普通名詞 1 * 0 * 0 "代表表記:天気/てんき カテゴリ:抽象物" <代表表記:天気/てんき><カテゴリ:抽象物><正規化代表表記:天気/てんき><漢字><かな漢字><名詞相当語><文頭><自立><内容語><タグ単位始><文節始><文節主辞>
            が が が 助詞 9 格助詞 1 * 0 * 0 NIL <かな漢字><ひらがな><付属>
            * 2D <BGH:良い/よい><時制:過去><用言:形><係:連用><レベル:B+><区切:3-5><ID:〜ので><提題受:20><連用要素><連用節><状態述語><正規化代表表記:良い/よい><主辞代表表記:良い/よい>
            + 2D <BGH:良い/よい><時制:過去><用言:形><係:連用><レベル:B+><区切:3-5><ID:〜ので><提題受:20><連用要素><連用節><状態述語><節-機能-原因・理由:ので><正規化代表表記:良い/よい><主辞代表表記:良い/よい><用言代表表記:良い/よい><節-区切><節-主辞>
            良かった よかった 良い 形容詞 3 * 0 イ形容詞アウオ段 18 タ形 8 "代表表記:良い/よい 反義:形容詞:悪い/わるい" <代表表記:良い/よい><反義:形容詞:悪い/わるい><正規化代表表記:良い/よい><かな漢字><活用語><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            ので ので のだ 助動詞 5 * 0 ナ形容詞 21 ダ列タ系連用テ形 12 NIL <かな漢字><ひらがな><活用語><付属>
            * -1D <BGH:散歩/さんぽ+する/する><文末><サ変><サ変動詞><時制:過去><句点><用言:動><レベル:C><区切:5-5><ID:（文末）><係:文末><提題受:30><主節><格要素><連用要素><動態述語><正規化代表表記:散歩/さんぽ><主辞代表表記:散歩/さんぽ>
            + -1D <BGH:散歩/さんぽ+する/する><文末><サ変動詞><時制:過去><句点><用言:動><レベル:C><区切:5-5><ID:（文末）><係:文末><提題受:30><主節><格要素><連用要素><動態述語><サ変><正規化代表表記:散歩/さんぽ><主辞代表表記:散歩/さんぽ><用言代表表記:散歩/さんぽ><節-区切><節-主辞><主題格:一人称優位>
            散歩 さんぽ 散歩 名詞 6 サ変名詞 2 * 0 * 0 "代表表記:散歩/さんぽ ドメイン:レクリエーション カテゴリ:抽象物" <代表表記:散歩/さんぽ><ドメイン:レクリエーション><カテゴリ:抽象物><正規化代表表記:散歩/さんぽ><漢字><かな漢字><名詞相当語><サ変><サ変動詞><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            した した する 動詞 2 * 0 サ変動詞 16 タ形 10 "代表表記:する/する 自他動詞:自:成る/なる 付属動詞候補（基本）" <代表表記:する/する><自他動詞:自:成る/なる><付属動詞候補（基本）><正規化代表表記:する/する><かな漢字><ひらがな><活用語><表現文末><とタ系連用テ形複合辞><付属>
            。 。 。 特殊 1 句点 1 * 0 * 0 NIL <英記号><記号><文末><付属>
            EOS
            # S-ID: 2
            * 2D <SM-動作><BGH:途中/とちゅう><文頭><時間><デ><助詞><体言><修飾><係:デ格><区切:0-0><格要素><連用要素><正規化代表表記:途中/とちゅう><主辞代表表記:途中/とちゅう>
            + 2D <SM-動作><BGH:途中/とちゅう><文頭><時間><デ><助詞><体言><修飾><係:デ格><区切:0-0><格要素><連用要素><名詞項候補><正規化代表表記:途中/とちゅう><主辞代表表記:途中/とちゅう>
            途中 とちゅう 途中 名詞 6 時相名詞 10 * 0 * 0 "代表表記:途中/とちゅう カテゴリ:抽象物 弱時相名詞 修飾（デ格）" <代表表記:途中/とちゅう><カテゴリ:抽象物><弱時相名詞><修飾（デ格）><正規化代表表記:途中/とちゅう><漢字><かな漢字><名詞相当語><文頭><自立><内容語><タグ単位始><文節始><文節主辞>
            で で で 助詞 9 格助詞 1 * 0 * 0 NIL <かな漢字><ひらがな><付属>
            * 2D <SM-主体><SM-人><BGH:先生/せんせい><ニ><助詞><体言><係:ニ格><区切:0-0><格要素><連用要素><正規化代表表記:先生/せんせい><主辞代表表記:先生/せんせい>
            + 2D <SM-主体><SM-人><BGH:先生/せんせい><ニ><助詞><体言><係:ニ格><区切:0-0><格要素><連用要素><名詞項候補><先行詞候補><正規化代表表記:先生/せんせい><主辞代表表記:先生/せんせい>
            先生 せんせい 先生 名詞 6 普通名詞 1 * 0 * 0 "代表表記:先生/せんせい ドメイン:教育・学習 カテゴリ:人 人名末尾" <代表表記:先生/せんせい><ドメイン:教育・学習><カテゴリ:人><人名末尾><正規化代表表記:先生/せんせい><漢字><かな漢字><名詞相当語><自立><内容語><タグ単位始><文節始><文節主辞>
            に に に 助詞 9 格助詞 1 * 0 * 0 NIL <かな漢字><ひらがな><付属>
            * -1D <BGH:会う/あう><文末><時制:過去><句点><用言:動><レベル:C><区切:5-5><ID:（文末）><係:文末><提題受:30><主節><格要素><連用要素><動態述語><正規化代表表記:会う/あう><主辞代表表記:会う/あう>
            + -1D <BGH:会う/あう><文末><時制:過去><句点><用言:動><レベル:C><区切:5-5><ID:（文末）><係:文末><提題受:30><主節><格要素><連用要素><動態述語><正規化代表表記:会う/あう><主辞代表表記:会う/あう><用言代表表記:会う/あう><節-区切><節-主辞><主題格:一人称優位>
            会った あった 会う 動詞 2 * 0 子音動詞ワ行 12 タ形 10 "代表表記:会う/あう 反義:動詞:分かれる/わかれる;動詞:別れる/わかれる" <代表表記:会う/あう><反義:動詞:分かれる/わかれる;動詞:別れる/わかれる><正規化代表表記:会う/あう><かな漢字><活用語><表現文末><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            。 。 。 特殊 1 句点 1 * 0 * 0 NIL <英記号><記号><文末><付属>
            EOS
            \"\"\"
            doc = Document.from_knp(knp_text)

        .. note::
            複数文の解析結果が含まれている場合，一つの文書として扱われる．
        """
        document = cls()
        sentences = []
        sentence_lines: list[str] = []
        for line in knp_text.split("\n"):
            if line.strip() == "":
                continue
            sentence_lines.append(line)
            if line.strip() == Sentence.EOS:
                try:
                    sentences.append(
                        Sentence.from_knp("\n".join(sentence_lines) + "\n")
                    )
                except Exception as e:
                    document.has_error = True
                    if not ignore_errors:
                        raise e
                sentence_lines = []
        document.sentences = sentences
        document._post_init()
        return document

    def to_plain(self) -> str:
        """プレーンテキストフォーマットに変換．

        .. note::
            文分割済みの場合は一行一文の形式で出力．
        """
        if self.need_senter:
            return self.text.rstrip() + "\n"
        return "".join(sentence.to_plain() for sentence in self.sentences)

    def to_jumanpp(self) -> str:
        """Juman++ フォーマットに変換．"""
        return "".join(sentence.to_jumanpp() for sentence in self.sentences)

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        return "".join(sentence.to_knp() for sentence in self.sentences)

    def _parse_rel(self) -> None:
        """関係タグ付きコーパスにおける <rel> タグをパース．"""
        for base_phrase in self.base_phrases:
            pas = Pas(Predicate(base_phrase))
            for rel in base_phrase.rels:
                if rel.type not in ALL_CASES:
                    continue
                if rel.sid is not None:
                    sentence = next(
                        sent
                        for sent in base_phrase.document.sentences
                        if sent.sid == rel.sid
                    )
                    assert rel.base_phrase_index is not None
                    arg_base_phrase = sentence.base_phrases[rel.base_phrase_index]
                    assert rel.target in arg_base_phrase.text
                    pas.add_argument(rel.type, arg_base_phrase, mode=rel.mode)
                else:
                    pas.add_special_argument(
                        rel.type, rel.target, base_phrase.index, mode=rel.mode
                    )  # TODO: fix eid
            self._pass.append(pas)



from logging import getLogger
from subprocess import PIPE, Popen, run
from typing import Optional, Union

from rhoknp.processors.jumanpp import Jumanpp
from rhoknp.processors.processor import Processor
from rhoknp.processors.senter import RegexSenter
from rhoknp.units import Document, Sentence

logger = getLogger(__file__)


class KNP(Processor):
    """KNP クラス．

    Args:
        executable: KNP のパス．
        options: KNP のオプション．
        senter: 文分割器のインスタンス．文分割がまだなら，先にこのインスタンスを用いて文分割する．
            未設定なら RegexSenter を使って文分割する．
        jumanpp: Jumanpp のインスタンス．形態素解析がまだなら，先にこのインスタンスを用いて形態素解析する．
            未設定なら Jumanpp （オプションなし）を使って形態素解析する．

    Example::

        from rhoknp import KNP

        knp = KNP()
        sentence = knp.apply("電気抵抗率は、どんな材料が電気を通しにくいかを比較するために、用いられる物性値である。")
    """

    def __init__(
        self,
        executable: str = "knp",
        options: Optional[list[str]] = None,
        senter: Optional[Processor] = None,
        jumanpp: Optional[Processor] = None,
    ):
        self.executable = executable
        self.options = options
        self.senter = senter
        self.jumanpp = jumanpp

    def __repr__(self) -> str:
        arg_string = f"executable={repr(self.executable)}"
        if self.options is not None:
            arg_string += f", options={repr(self.options)}"
        if self.senter is not None:
            arg_string += f", senter={repr(self.senter)}"
        if self.jumanpp is not None:
            arg_string += f", jumanpp={repr(self.jumanpp)}"
        return f"{self.__class__.__name__}({arg_string})"

    def apply_to_document(self, document: Union[Document, str]) -> Document:
        """文書に KNP を適用する．

        Args:
            document: 文書．

        .. note::
            文分割がまだなら，先に初期化時に設定した senter で文分割する．
            未設定なら RegexSenter で文分割する．
            形態素解析がまだなら，先に初期化時に設定した jumanpp で形態素解析する．
            未設定なら Jumanpp （オプションなし）で形態素解析する．
        """
        if isinstance(document, str):
            document = Document(document)

        if document.need_senter:
            logger.debug("document needs to be split into sentences")
            if self.senter is None:
                logger.debug(
                    "senter is not specified when initializing KNP: use RegexSenter with no option"
                )
                self.senter = RegexSenter()
            document = self.senter.apply_to_document(document)

        if document.need_jumanpp:
            logger.debug("document needs to be processed by Juman++")
            if self.jumanpp is None:
                logger.info(
                    "jumanpp is not specified when initializing KNP: use Jumanpp with no option"
                )
                self.jumanpp = Jumanpp()
            document = self.jumanpp.apply_to_document(document)

        with Popen(self.run_command, stdout=PIPE, stdin=PIPE, encoding="utf-8") as p:
            knp_text, _ = p.communicate(input=document.to_jumanpp())
        return Document.from_knp(knp_text)

    def apply_to_sentence(self, sentence: Union[Sentence, str]) -> Sentence:
        """文に KNP を適用する．

        Args:
            sentence: 文．

        .. note::
            形態素解析がまだなら，先に初期化時に設定した jumanpp で形態素解析する．
            未設定なら Jumanpp （オプションなし）で形態素解析する．
        """
        if isinstance(sentence, str):
            sentence = Sentence(sentence)

        if sentence.need_jumanpp:
            logger.debug("sentence needs to be processed by Juman++")
            if self.jumanpp is None:
                logger.info(
                    "jumanpp is not specified when initializing KNP: use Jumanpp with no option"
                )
                self.jumanpp = Jumanpp()
            sentence = self.jumanpp.apply_to_sentence(sentence)

        with Popen(self.run_command, stdout=PIPE, stdin=PIPE, encoding="utf-8") as p:
            knp_text, _ = p.communicate(input=sentence.to_jumanpp())
        return Sentence.from_knp(knp_text)

    def is_available(self) -> bool:
        """KNP が利用可能であれば True を返す．"""
        try:
            p = run(self.version_command, stdout=PIPE, stdin=PIPE, encoding="utf-8")
            logger.info(p.stdout.strip())
            return True
        except Exception as e:
            logger.warning(e)
            return False

    @property
    def run_command(self) -> list[str]:
        """解析時に実行するコマンド．"""
        command = [self.executable]
        if self.options:
            command += self.options
        else:
            command += ["-tab"]
        return command

    @property
    def version_command(self) -> list[str]:
        """バージョン確認時に実行するコマンド．"""
        return [self.executable, "-v"]



from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.clause import Clause
from rhoknp.units.document import Document
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.phrase import Phrase
from rhoknp.units.sentence import Sentence

__all__ = ["Document", "Sentence", "Clause", "Phrase", "BasePhrase", "Morpheme"]



import re
import weakref
from functools import cached_property
from typing import TYPE_CHECKING, Optional, Union

from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.unit import Unit
from rhoknp.units.utils import DepType, Features

if TYPE_CHECKING:
    from rhoknp.units.clause import Clause
    from rhoknp.units.document import Document
    from rhoknp.units.sentence import Sentence


class Phrase(Unit):
    """文節クラス．"""

    KNP_PAT = re.compile(
        rf"^\* (?P<pid>-1|\d+)(?P<dtype>[DPAI])( {Features.PAT.pattern})?$"
    )
    count = 0

    def __init__(self, parent_index: int, dep_type: DepType, features: Features):
        super().__init__()

        # parent unit
        self._clause: Optional["Clause"] = None
        self._sentence: Optional["Sentence"] = None

        # child units
        self._base_phrases: Optional[list[BasePhrase]] = None

        self.parent_index: int = parent_index  #: 係り先の文節の文内におけるインデックス．
        self.dep_type: DepType = dep_type  #: 係り受けの種類．
        self.features: Features = features  #: 素性．

        self.index = self.count
        Phrase.count += 1

    @property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        offset = 0
        for prev_sentence in self.document.sentences[: self.sentence.index]:
            offset += len(prev_sentence.phrases)
        return self.index + offset

    @property
    def parent_unit(self) -> Optional[Union["Clause", "Sentence"]]:
        """上位の言語単位（節もしくは文）．未登録なら None．"""
        if self._clause is not None:
            return self._clause
        if self._sentence is not None:
            return self._sentence
        return None

    @property
    def child_units(self) -> Optional[list[BasePhrase]]:
        """下位の言語単位（基本句）．解析結果にアクセスできないなら None．"""
        return self._base_phrases

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self.sentence.document

    @property
    def sentence(self) -> "Sentence":
        """文．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return self._sentence or self.clause.sentence

    @sentence.setter
    def sentence(self, sentence: "Sentence") -> None:
        """文．

        Args:
            sentence: 文．
        """
        self._sentence = sentence

    @property
    def clause(self) -> "Clause":
        """節．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._clause is None:
            raise AttributeError("clause has not been set")
        return self._clause

    @clause.setter
    def clause(self, clause: "Clause") -> None:
        """節．

        Args:
            clause: 節．
        """
        self._clause = clause

    @property
    def base_phrases(self) -> list[BasePhrase]:
        """基本句のリスト．"""
        if self._base_phrases is None:
            raise AssertionError
        return self._base_phrases

    @base_phrases.setter
    def base_phrases(self, base_phrases: list[BasePhrase]) -> None:
        """基本句．

        Args:
            base_phrases: 基本句．
        """
        for base_phrase in base_phrases:
            base_phrase.phrase = weakref.proxy(self)
        self._base_phrases = base_phrases

    @property
    def morphemes(self) -> list[Morpheme]:
        """形態素のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [
            morpheme
            for base_phrase in self.base_phrases
            for morpheme in base_phrase.morphemes
        ]

    @property
    def parent(self) -> Optional["Phrase"]:
        """係り先の文節．ないなら None．"""
        if self.parent_index == -1:
            return None
        return self.sentence.phrases[self.parent_index]

    @cached_property
    def children(self) -> list["Phrase"]:
        """この文節に係っている文節のリスト．"""
        return [phrase for phrase in self.sentence.phrases if phrase.parent == self]

    @classmethod
    def from_knp(cls, knp_text: str) -> "Phrase":
        """文節クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．
        """
        first_line, *lines = knp_text.split("\n")
        match = cls.KNP_PAT.match(first_line)
        if match is None:
            raise ValueError(f"malformed line: {first_line}")
        parent_index = int(match.group("pid"))
        dep_type = DepType(match.group("dtype"))
        features = Features(match.group("feats") or "")
        phrase = cls(parent_index, dep_type, features)

        base_phrases: list[BasePhrase] = []
        base_phrase_lines: list[str] = []
        for line in lines:
            if not line.strip():
                continue
            if line.startswith("+") and base_phrase_lines:
                base_phrases.append(BasePhrase.from_knp("\n".join(base_phrase_lines)))
                base_phrase_lines = []
            base_phrase_lines.append(line)
        else:
            base_phrases.append(BasePhrase.from_knp("\n".join(base_phrase_lines)))
        phrase.base_phrases = base_phrases
        return phrase

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        ret = f"* {self.parent_index}{self.dep_type.value}"
        if self.features:
            ret += f" {self.features}"
        ret += "\n"
        ret += "".join(base_phrase.to_knp() for base_phrase in self.base_phrases)
        return ret



from abc import ABC, abstractmethod
from typing import Any, Optional, Sequence


class Unit(ABC):
    def __init__(self) -> None:
        self._index: Optional[int] = None

        self._text: Optional[str] = None

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(index={repr(self.index)}, text={repr(self.text)})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, type(self)):
            if self.parent_unit != other.parent_unit:
                return False
            return self.index == other.index
        return False

    @property
    def parent_unit(self) -> Optional["Unit"]:
        raise NotImplementedError

    @property
    @abstractmethod
    def child_units(self) -> Optional[Sequence["Unit"]]:
        raise NotImplementedError

    @property
    def index(self) -> int:
        assert self._index is not None
        return self._index

    @index.setter
    def index(self, index: int) -> None:
        self._index = index

    @property
    def text(self) -> str:
        if self.child_units is not None:
            return "".join(str(child_unit) for child_unit in self.child_units)
        elif self._text is not None:
            return self._text
        raise AssertionError

    @text.setter
    def text(self, text: str) -> None:
        self._text = text



ALL_CASES = [
    "ガ",
    "デ",
    "ト",
    "ニ",
    "ノ",
    "ヘ",
    "ヲ",
    "カラ",
    "ガ２",
    "ノ？",
    "マデ",
    "ヨリ",
    "トイウ",
    "トシテ",
    "トスル",
    "ニオク",
    "ニシテ",
    "ニツク",
    "ニトル",
    "ニヨル",
    "マデニ",
    "ニオイテ",
    "ニカワル",
    "ニソッテ",
    "ニツイテ",
    "ニトッテ",
    "ニムケテ",
    "ニムケル",
    "ニヨッテ",
    "ニヨラズ",
    "ニアワセテ",
    "ニカギッテ",
    "ニカギラズ",
    "ニカランデ",
    "ニカワッテ",
    "ニカンシテ",
    "ニカンスル",
    "ニクラベテ",
    "ニクワエテ",
    "ニタイシテ",
    "ニタイスル",
    "ニツヅイテ",
    "ニナランデ",
    "ヲツウジテ",
    "ヲツウジル",
    "ヲノゾイテ",
    "ヲフクメテ",
    "ヲメグッテ",
    "ニトモナッテ",
    "ニモトヅイテ",
    "無",
    "修飾",
    "判ガ",
    "時間",
    "外の関係",
]
ALL_CASES += [case + "≒" for case in ALL_CASES]

ALL_COREFS = ["=", "=構", "=役"]
ALL_COREFS += [case + "≒" for case in ALL_COREFS]



import re
import sys
from collections import defaultdict
from enum import Enum, auto
from typing import Optional

from rhoknp.pas.argument import Argument, ArgumentType, BaseArgument, SpecialArgument
from rhoknp.pas.predicate import Predicate
from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.utils import RelMode


class CaseInfoFormat(Enum):
    CASE = auto()  # 格解析フォーマット
    PAS = auto()  # 述語項構造フォーマット


class Pas:
    ARGUMENT_PAT = re.compile(
        r"([^/;]+/[CNODEU-]/[^/]+/(-?\d*)/(-?\d*)/[^/;]+)"
    )  # ガ/N/彼/0/0/5

    def __init__(self, predicate: Predicate):
        self._predicate = predicate
        predicate.pas = self
        self._arguments: dict[str, list[BaseArgument]] = defaultdict(list)
        self.modes: dict[str, RelMode] = {}

    @property
    def predicate(self) -> Predicate:
        return self._predicate

    @property
    def arguments(self) -> dict[str, list[BaseArgument]]:
        return self._arguments

    @classmethod
    def _from_pas_string(
        cls, base_phrase: BasePhrase, fstring: str, format_: CaseInfoFormat
    ) -> "Pas":
        # language=RegExp
        cfid_pat = r"(.*?):([^:/]+?)"  # 食べる/たべる:動1
        match = re.match(
            r"{cfid}(:(?P<args>{args}(;{args})*))?$".format(
                cfid=cfid_pat, args=cls.ARGUMENT_PAT.pattern
            ),
            fstring,
        )

        if match is None:
            print(f"invalid tag format: '{fstring}' is ignored", file=sys.stderr)
            return cls(Predicate(base_phrase))

        cfid = match.group(1) + ":" + match.group(2)
        predicate = Predicate(unit=base_phrase, cfid=cfid)

        if match.group(3) is None:  # <述語項構造:束の間/つかのま:判0> など
            return cls(predicate)

        pas = cls(predicate)
        for match_arg in cls.ARGUMENT_PAT.finditer(match.group("args")):
            case, case_flag, surf, *fields = match_arg.group(0).split("/")
            if case_flag in ("U", "-"):
                continue
            arg_type = ArgumentType(case_flag)
            arg: BaseArgument
            if format_ == CaseInfoFormat.CASE:
                tid, sdist, sid = int(fields[0]), int(fields[1]), fields[2]
                assert arg_type != ArgumentType.EXOPHOR
                assert (
                    base_phrase.document.sentences[
                        base_phrase.sentence.index - sdist
                    ].sid
                    == sid
                )
                arg_base_phrase = base_phrase.document.sentences[
                    base_phrase.sentence.index - sdist
                ].base_phrases[tid]
                assert surf in arg_base_phrase.text
                pas.add_argument(case, arg_base_phrase, arg_type=arg_type)
            else:
                assert format_ == CaseInfoFormat.PAS
                sdist, tid, eid = int(fields[0]), int(fields[1]), int(fields[2])
                if arg_type == ArgumentType.EXOPHOR:
                    pas.add_special_argument(case, surf, eid)
                else:
                    arg_base_phrase = base_phrase.document.sentences[
                        base_phrase.sentence.index - sdist
                    ].base_phrases[tid]
                    assert surf in arg_base_phrase.text
                    pas.add_argument(case, arg_base_phrase)
        return pas

    @classmethod
    def from_base_phrase(cls, base_phrase: BasePhrase) -> "Pas":
        if "述語項構造" in base_phrase.features:
            pas_string = base_phrase.features["述語項構造"]
            assert isinstance(pas_string, str)
            pas = cls._from_pas_string(
                base_phrase, pas_string, format_=CaseInfoFormat.PAS
            )
        elif "格解析結果" in base_phrase.features:
            pas_string = base_phrase.features["格解析結果"]
            assert isinstance(pas_string, str)
            pas = cls._from_pas_string(
                base_phrase, pas_string, format_=CaseInfoFormat.CASE
            )
        else:
            pas = cls(Predicate(base_phrase))
        base_phrase.pas = pas
        return pas

    def add_argument(
        self,
        case: str,
        base_phrase: BasePhrase,
        mode: Optional[RelMode] = None,
        arg_type: Optional[ArgumentType] = None,
    ) -> None:
        argument = Argument(
            base_phrase,
            arg_type or self._get_arg_type(self.predicate, base_phrase, case),
        )
        argument.pas = self
        if mode is not None:
            self.modes[case] = mode
        if argument not in self.arguments[case]:
            self.arguments[case].append(argument)

    def add_special_argument(
        self, case: str, exophor: str, eid: int, mode: Optional[RelMode] = None
    ) -> None:
        special_argument = SpecialArgument(exophor, eid)
        special_argument.pas = self
        if mode is not None:
            self.modes[case] = mode
        if special_argument not in self.arguments[case]:
            self.arguments[case].append(special_argument)

    @staticmethod
    def _get_arg_type(
        predicate: Predicate, arg_base_phrase: BasePhrase, case: str
    ) -> ArgumentType:
        if arg_base_phrase in predicate.phrase.children:
            dep_case = arg_base_phrase.features.get("係", "")
            assert isinstance(dep_case, str)
            dep_case = dep_case.rstrip("格")
            if (
                (case == dep_case)
                or (case == "判ガ" and dep_case == "ガ")
                or (case == "ノ？" and dep_case == "ノ")
            ):
                return ArgumentType.CASE_EXPLICIT
            else:
                return ArgumentType.CASE_HIDDEN
        elif predicate.phrase.parent and predicate.phrase.parent == arg_base_phrase:
            return ArgumentType.CASE_HIDDEN
        else:
            return ArgumentType.OMISSION

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(predicate={repr(self.predicate)}, arguments={repr(dict(self.arguments))})"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rhoknp', 'rhoknp.pas', 'rhoknp.processors', 'rhoknp.units', 'rhoknp.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'rhoknp',
    'version': '0.1.1',
    'description': 'Yet another Python binding for Juman++/KNP',
    'long_description': None,
    'author': 'Hirokazu Kiyomaru',
    'author_email': 'h.kiyomaru@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ku-nlp/rhoknp',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import re
from enum import Enum
from typing import Optional, ClassVar, AnyStr


class ExophoraReferentType(Enum):
    WRITER = "著者"
    READER = "読者"
    UNSPECIFIED_PERSON = "不特定:人"
    UNSPECIFIED_MATTER = "不特定:物"
    UNSPECIFIED_SITUATION = "不特定:状況"
    PREVIOUS_SENTENCE = "前文"
    NEXT_SENTENCE = "後文"


class ExophoraReferent:
    _NUM2ZEN: dict[int, str] = {0: "０", 1: "１", 2: "２", 3: "３", 4: "４", 5: "５", 6: "６", 7: "７", 8: "８", 9: "９"}
    PAT: ClassVar[re.Pattern[str]] = re.compile(
        rf"^({'|'.join(t.value for t in ExophoraReferentType)})(?P<index>[0-9０-９]*)$"
    )

    def __init__(self, text: str) -> None:
        match: Optional[re.Match[AnyStr]] = self.PAT.match(text)
        if match is None:
            raise ValueError(f"invalid exophora referent: {text}")
        self.index: Optional[str] = match.group("index")
        self.type = ExophoraReferentType(text)

    @property
    def text(self) -> str:
        return self.type.value + "".join(self._NUM2ZEN[int(n)] for n in self.index)

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text={repr(self.text)})"



import re
import weakref
from typing import TYPE_CHECKING, Optional, Union

from rhoknp.units.base_phrase import BasePhrase
from rhoknp.units.clause import Clause
from rhoknp.units.morpheme import Morpheme
from rhoknp.units.phrase import Phrase
from rhoknp.units.unit import Unit
from rhoknp.units.utils import is_comment_line

if TYPE_CHECKING:
    from rhoknp.units.document import Document


class Sentence(Unit):
    """文クラス．

    Args:
        text: 文の文字列．
    """

    EOS = "EOS"
    SID_PAT = re.compile(r"^(?P<sid>(?P<did>[a-zA-Z0-9-_]+?)(-(\d+))?)$")
    SID_PAT_KWDLC = re.compile(r"^(?P<sid>(?P<did>w\d{6}-\d{10})(-\d+){1,2})$")
    SID_PAT_WAC = re.compile(r"^(?P<sid>(?P<did>wiki\d{8})(-\d{2})(-\d{2})?)$")
    count = 0

    def __init__(self, text: Optional[str] = None):
        super().__init__()
        if text is not None:
            self.text = text

        Clause.count = 0
        Phrase.count = 0
        BasePhrase.count = 0
        Morpheme.count = 0

        # parent unit
        self._document: Optional["Document"] = None

        # child units
        self._clauses: Optional[list[Clause]] = None
        self._phrases: Optional[list[Phrase]] = None
        self._morphemes: Optional[list[Morpheme]] = None

        self.sid: Optional[str] = None
        self.misc_comment: str = ""

        self.index = self.count
        Sentence.count += 1

    @property
    def global_index(self) -> int:
        """文書全体におけるインデックス．"""
        return self.index

    @property
    def parent_unit(self) -> Optional["Document"]:
        """上位の言語単位（文書）．未登録なら None．"""
        return self._document

    @property
    def child_units(
        self,
    ) -> Optional[Union[list[Clause], list[Phrase], list[Morpheme]]]:
        """下位の言語単位（節もしくは形態素）のリスト．解析結果にアクセスできないなら None．

        .. note::
            KNP によって解析済みなら節， Jumanpp によって解析済みなら形態素のリストを返却．
            KNP による素性が付与されていない場合は節境界が判断できないため文節を返却．
        """
        if self._clauses is not None:
            return self._clauses
        elif self._phrases is not None:
            return self._phrases
        elif self._morphemes is not None:
            return self._morphemes
        return None

    @property
    def document(self) -> "Document":
        """文書．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self.parent_unit is None:
            raise AttributeError("document has not been set")
        return self.parent_unit

    @document.setter
    def document(self, document: "Document") -> None:
        """文書．

        Args:
            document: 文書．
        """
        self._document = document

    @property
    def clauses(self) -> list[Clause]:
        """節のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._clauses is None:
            raise AttributeError("not available before applying KNP")
        return self._clauses

    @clauses.setter
    def clauses(self, clauses: list[Clause]) -> None:
        """節のリスト．

        Args:
            clauses: 節のリスト．
        """
        for clause in clauses:
            clause.sentence = weakref.proxy(self)
        self._clauses = clauses

    @property
    def phrases(self) -> list[Phrase]:
        """文節のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._phrases is not None:
            return self._phrases
        elif self._clauses is not None:
            return [phrase for clause in self.clauses for phrase in clause.phrases]
        raise AttributeError("not available before applying KNP")

    @phrases.setter
    def phrases(self, phrases: list[Phrase]) -> None:
        """文節のリスト．

        Args:
            phrases: 文節のリスト．
        """
        for phrase in phrases:
            phrase.sentence = weakref.proxy(self)
        self._phrases = phrases

    @property
    def base_phrases(self) -> list[BasePhrase]:
        """基本句のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        return [
            base_phrase
            for phrase in self.phrases
            for base_phrase in phrase.base_phrases
        ]

    @property
    def morphemes(self) -> list[Morpheme]:
        """形態素のリスト．

        Raises:
            AttributeError: 解析結果にアクセスできない場合．
        """
        if self._morphemes is not None:
            return self._morphemes
        elif self._clauses is not None:
            return [
                morpheme
                for base_phrase in self.base_phrases
                for morpheme in base_phrase.morphemes
            ]
        raise AttributeError("not available before applying Jumanpp")

    @morphemes.setter
    def morphemes(self, morphemes: list[Morpheme]) -> None:
        """形態素のリスト．

        Args:
            morphemes: 形態素のリスト．
        """
        for morpheme in morphemes:
            morpheme.sentence = weakref.proxy(self)
        self._morphemes = morphemes

    @property
    def comment(self) -> str:
        """コメント行．"""
        ret = ""
        if sid := self.sid:
            ret += f"S-ID:{sid} "
        if misc := self.misc_comment:
            ret += f"{misc} "
        if ret != "":
            ret = "# " + ret
        return ret.rstrip(" ")

    @comment.setter
    def comment(self, comment: str) -> None:
        """コメント行．

        Args:
            comment: コメント行．
        """
        doc_id, sid, rest = self._extract_sid(comment)
        if sid is not None:
            self.sid = sid
        self.misc_comment = rest

    @property
    def need_jumanpp(self) -> bool:
        """Juman++ による形態素解析がまだなら True．"""
        return (
            self._morphemes is None and self._phrases is None and self._clauses is None
        )

    @property
    def need_knp(self) -> bool:
        """KNP による構文解析がまだなら True．"""
        return self.need_jumanpp or self._phrases is None and self._clauses is None

    @classmethod
    def from_raw_text(cls, text: str) -> "Sentence":
        """文クラスのインスタンスを文の文字列から初期化．

        Args:
            text: 文の文字列．

        Example::

            from rhoknp import Sentence

            sent_text = "天気が良かったので散歩した。"
            sent = Sentence(sent_text)
        """
        sentence = cls()
        text_lines = []
        for line in text.split("\n"):
            if is_comment_line(line):
                sentence.comment = line
            else:
                text_lines.append(line)
        sentence.text = "\n".join(text_lines)
        return sentence

    @classmethod
    def from_jumanpp(cls, jumanpp_text: str) -> "Sentence":
        """文クラスのインスタンスを Juman++ の解析結果から初期化．

        Args:
            jumanpp_text: Juman++ の解析結果．

        Example::

            from rhoknp import Sentence

            jumanpp_text = \"\"\"天気 てんき 天気 名詞 6 普通名詞 1 * 0 * 0 "代表表記:天気/てんき カテゴリ:抽象物" <代表表記:天気/てんき><カテゴリ:抽象物><正規化代表表記:天気/てんき><漢字><かな漢字><名詞相当語><文頭><自立><内容語><タグ単位始><文節始><文節主辞>
            が が が 助詞 9 格助詞 1 * 0 * 0 NIL <かな漢字><ひらがな><付属>
            良かった よかった 良い 形容詞 3 * 0 イ形容詞アウオ段 18 タ形 8 "代表表記:良い/よい 反義:形容詞:悪い/わるい"
            ので ので のだ 助動詞 5 * 0 ナ形容詞 21 ダ列タ系連用テ形 12 NIL <かな漢字><ひらがな><活用語><付属>
            散歩 さんぽ 散歩 名詞 6 サ変名詞 2 * 0 * 0 "代表表記:散歩/さんぽ ドメイン:レクリエーション カテゴリ:抽象物" <代表表記:散歩/さんぽ><ドメイン:レクリエーション><カテゴリ:抽象物><正規化代表表記:散歩/さんぽ><漢字><かな漢字><名詞相当語><サ変><サ変動詞><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            した した する 動詞 2 * 0 サ変動詞 16 タ形 10 "代表表記:する/する 自他動詞:自:成る/なる 付属動詞候補（基本）" <代表表記:する/する><自他動詞:自:成る/なる><付属動詞候補（基本）><正規化代表表記:する/する><かな漢字><ひらがな><活用語><表現文末><とタ系連用テ形複合辞><付属>
            。 。 。 特殊 1 句点 1 * 0 * 0 NIL <英記号><記号><文末><付属>
            EOS
            \"\"\"
            sent = Sentence.from_jumanpp(jumanpp_text)
        """
        sentence = cls()
        morphemes: list[Morpheme] = []
        jumanpp_lines: list[str] = []
        for line in jumanpp_text.split("\n"):
            if not line.strip():
                continue
            if is_comment_line(line):
                sentence.comment = line
                continue
            elif line.startswith("@") and not line.startswith("@ @"):
                # homograph
                pass
            elif jumanpp_lines:
                morphemes.append(Morpheme.from_jumanpp("\n".join(jumanpp_lines)))
                jumanpp_lines = []
            jumanpp_lines.append(line)
            if line.strip() == cls.EOS:
                break
        sentence.morphemes = morphemes
        return sentence

    @classmethod
    def from_knp(cls, knp_text: str) -> "Sentence":
        """文クラスのインスタンスを KNP の解析結果から初期化．

        Args:
            knp_text: KNP の解析結果．

        Example::

            from rhoknp import Sentence

            knp_text = \"\"\"# S-ID: 1
            * 1D <BGH:天気/てんき><文頭><ガ><助詞><体言><係:ガ格><区切:0-0><格要素><連用要素><正規化代表表記:天気/てんき><主辞代表表記:天気/てんき>
            + 1D <BGH:天気/てんき><文頭><ガ><助詞><体言><係:ガ格><区切:0-0><格要素><連用要素><名詞項候補><先行詞候補><正規化代表表記:天気/てんき><主辞代表表記:天気/てんき>
            天気 てんき 天気 名詞 6 普通名詞 1 * 0 * 0 "代表表記:天気/てんき カテゴリ:抽象物" <代表表記:天気/てんき><カテゴリ:抽象物><正規化代表表記:天気/てんき><漢字><かな漢字><名詞相当語><文頭><自立><内容語><タグ単位始><文節始><文節主辞>
            が が が 助詞 9 格助詞 1 * 0 * 0 NIL <かな漢字><ひらがな><付属>
            * 2D <BGH:良い/よい><時制:過去><用言:形><係:連用><レベル:B+><区切:3-5><ID:〜ので><提題受:20><連用要素><連用節><状態述語><正規化代表表記:良い/よい><主辞代表表記:良い/よい>
            + 2D <BGH:良い/よい><時制:過去><用言:形><係:連用><レベル:B+><区切:3-5><ID:〜ので><提題受:20><連用要素><連用節><状態述語><節-機能-原因・理由:ので><正規化代表表記:良い/よい><主辞代表表記:良い/よい><用言代表表記:良い/よい><節-区切><節-主辞>
            良かった よかった 良い 形容詞 3 * 0 イ形容詞アウオ段 18 タ形 8 "代表表記:良い/よい 反義:形容詞:悪い/わるい" <代表表記:良い/よい><反義:形容詞:悪い/わるい><正規化代表表記:良い/よい><かな漢字><活用語><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            ので ので のだ 助動詞 5 * 0 ナ形容詞 21 ダ列タ系連用テ形 12 NIL <かな漢字><ひらがな><活用語><付属>
            * -1D <BGH:散歩/さんぽ+する/する><文末><サ変><サ変動詞><時制:過去><句点><用言:動><レベル:C><区切:5-5><ID:（文末）><係:文末><提題受:30><主節><格要素><連用要素><動態述語><正規化代表表記:散歩/さんぽ><主辞代表表記:散歩/さんぽ>
            + -1D <BGH:散歩/さんぽ+する/する><文末><サ変動詞><時制:過去><句点><用言:動><レベル:C><区切:5-5><ID:（文末）><係:文末><提題受:30><主節><格要素><連用要素><動態述語><サ変><正規化代表表記:散歩/さんぽ><主辞代表表記:散歩/さんぽ><用言代表表記:散歩/さんぽ><節-区切><節-主辞><主題格:一人称優位>
            散歩 さんぽ 散歩 名詞 6 サ変名詞 2 * 0 * 0 "代表表記:散歩/さんぽ ドメイン:レクリエーション カテゴリ:抽象物" <代表表記:散歩/さんぽ><ドメイン:レクリエーション><カテゴリ:抽象物><正規化代表表記:散歩/さんぽ><漢字><かな漢字><名詞相当語><サ変><サ変動詞><自立><内容語><タグ単位始><文節始><文節主辞><用言表記先頭><用言表記末尾><用言意味表記末尾>
            した した する 動詞 2 * 0 サ変動詞 16 タ形 10 "代表表記:する/する 自他動詞:自:成る/なる 付属動詞候補（基本）" <代表表記:する/する><自他動詞:自:成る/なる><付属動詞候補（基本）><正規化代表表記:する/する><かな漢字><ひらがな><活用語><表現文末><とタ系連用テ形複合辞><付属>
            。 。 。 特殊 1 句点 1 * 0 * 0 NIL <英記号><記号><文末><付属>
            EOS
            \"\"\"
            sent = Sentence.from_knp(knp_text)
        """
        lines = knp_text.split("\n")
        sentence = cls()
        has_clause_boundary = any(
            "節-区切" in line for line in lines if line.startswith("+")
        )
        clauses: list[Clause] = []
        phrases: list[Phrase] = []
        child_lines: list[str] = []
        is_clause_end = False
        for line in lines:
            if not line.strip():
                continue
            if is_comment_line(line):
                sentence.comment = line
                continue
            if line.startswith(";;"):
                raise Exception(f"Error: {line}")
            if line.startswith("+") and "節-区切" in line:
                is_clause_end = True
            if line.strip() == cls.EOS:
                if has_clause_boundary is True:
                    clauses.append(Clause.from_knp("\n".join(child_lines)))
                else:
                    phrases.append(Phrase.from_knp("\n".join(child_lines)))
                break
            if line.startswith("*"):
                if is_clause_end is True:
                    clauses.append(Clause.from_knp("\n".join(child_lines)))
                    child_lines = []
                    is_clause_end = False
                elif has_clause_boundary is False and child_lines:
                    phrases.append(Phrase.from_knp("\n".join(child_lines)))
                    child_lines = []
            child_lines.append(line)
        if has_clause_boundary is True:
            sentence.clauses = clauses
        else:
            sentence.phrases = phrases
        return sentence

    @staticmethod
    def _extract_sid(comment: str) -> tuple[Optional[str], Optional[str], str]:
        """Extract sentence id and document id from comment line.

        Args:
            comment: A comment line.

        Returns:
            Optional[str]: Document id if exists; otherwise, None.
            Optional[str]: Sentence id if exists; otherwise, None.
            str: The rest of the comment line.
        """
        if match_sid := re.match(r"# S-ID: ?(\S*)( .+)?$", comment):
            sid_string = match_sid.group(1)
            match = (
                Sentence.SID_PAT_KWDLC.match(sid_string)
                or Sentence.SID_PAT_WAC.match(sid_string)
                or Sentence.SID_PAT.match(sid_string)
            )
            if match is None:
                raise ValueError(f"unsupported S-ID format: {sid_string}")
            return (
                match.group("did"),
                match.group("sid"),
                match_sid.group(2).lstrip() if match_sid.group(2) else "",
            )
        assert comment.startswith("# ")
        return None, None, comment[2:]

    def to_plain(self) -> str:
        """プレーンテキストフォーマットに変換．"""
        ret = ""
        if self.comment != "":
            ret += self.comment + "\n"
        ret += self.text.rstrip("\n") + "\n"
        return ret

    def to_jumanpp(self) -> str:
        """Juman++ フォーマットに変換．"""
        ret = ""
        if self.comment != "":
            ret += self.comment + "\n"
        ret += (
            "".join(morpheme.to_jumanpp() for morpheme in self.morphemes)
            + self.EOS
            + "\n"
        )
        return ret

    def to_knp(self) -> str:
        """KNP フォーマットに変換．"""
        ret = ""
        if self.comment != "":
            ret += self.comment + "\n"
        ret += "".join(child.to_knp() for child in self._clauses or self.phrases)
        ret += self.EOS + "\n"
        return ret



import re
from logging import getLogger
from typing import Union

from rhoknp.processors.processor import Processor
from rhoknp.units import Document, Sentence

logger = getLogger(__file__)


class RegexSenter(Processor):
    """正規表現にもとづく文分割クラス．

    Example::
        from rhoknp import RegexSenter

        senter = RegexSenter()
        document = senter.apply_to_document("天気が良かったので散歩した。途中で先生に会った。")
    """

    PERIODS = "。．？！♪☆★…?!"

    def apply_to_document(self, document: Union[Document, str]) -> Document:
        if isinstance(document, str):
            document = Document(document)
        sentence_texts = self._split_document(document.text)
        return Document.from_sentences(sentence_texts)

    def apply_to_sentence(self, sentence: Union[Sentence, str]) -> Sentence:
        if isinstance(sentence, str):
            sentence = Sentence(sentence)
        return sentence

    def _split_document(self, text: str) -> list[str]:
        """Split text into sentences by regular expressions."""
        base = f"[^{self.PERIODS}]*[f{self.PERIODS}]"
        eol = f"[^{self.PERIODS}]*$"
        regex = re.compile(f"{base}|{eol}$")
        candidates = []
        for line in text.split("\n"):
            candidates += re.findall(regex, line + "\n")
        candidates = self._merge_candidates(candidates)
        return self._clean_up_candidates(candidates)

    def _merge_candidates(self, candidates: list[str]) -> list[str]:
        """Merge sentence candidates."""
        candidates = self._merge_single_periods(candidates)
        candidates = self._merge_parenthesis(candidates)
        return candidates

    def _merge_single_periods(self, candidates: list[str]) -> list[str]:
        """Merge sentence candidates that consist of just a single period."""
        regex = re.compile(f"^[{self.PERIODS}]$")
        merged_candidates = [""]
        for candidate in candidates:
            if re.match(regex, candidate):
                merged_candidates[-1] += candidate
            else:
                merged_candidates.append(candidate)
        if merged_candidates[0] == "":
            merged_candidates.pop(0)  # remove the dummy sentence
        return merged_candidates

    @staticmethod
    def _merge_parenthesis(sentence_candidates: list[str]) -> list[str]:
        """Merge sentence candidates so that strings in parentheses or
        brackets are not split.
        """
        parenthesis_level = 0
        quotation_level = 0

        merged_candidates = []
        prefix = ""
        while sentence_candidates:
            candidate = sentence_candidates.pop(0)

            parenthesis_level += candidate.count("（") + candidate.count("(")
            parenthesis_level -= candidate.count("）") + candidate.count(")")

            quotation_level += candidate.count("「") + candidate.count("“")
            quotation_level -= candidate.count("」") + candidate.count("”")

            if parenthesis_level == 0 and quotation_level == 0:
                candidate = prefix + candidate
                merged_candidates.append(candidate)
                prefix = ""
            else:
                if "\n" in candidate:
                    candidate, rest = candidate.split("\n", maxsplit=1)
                    candidate = prefix + candidate
                    merged_candidates.append(candidate)
                    prefix = ""
                    sentence_candidates.insert(0, rest)
                    parenthesis_level = 0
                    quotation_level = 0
                else:
                    prefix += candidate

        if prefix:
            merged_candidates.append(prefix)
        return merged_candidates

    @staticmethod
    def _clean_up_candidates(sentence_candidates: list[str]) -> list[str]:
        """Remove empty sentence candidates."""
        return [
            sentence_candidate.strip()
            for sentence_candidate in sentence_candidates
            if sentence_candidate.strip()
        ]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"
